from ursinanetworking import *
from easyursinanetworking import *

Server = UrsinaNetworkingServer("localhost", 25565)
Easy = EasyUrsinaNetworkingServer(Server)

Easy.create_replicated_variable("MyVariable", {"name" : "kevin"})
Easy.update_replicated_variable_by_name("MyVariablee", "name", "jean")
Easy.remove_replicated_variable_by_name("MyVariablee")

while True:
    Easy.process_net_events()


from ursina import *
import inspect
import uuid
import traceback

from ursinanetworking.ursinanetworking import UrsinaNetworkingClient, UrsinaNetworkingServer

def replicated_generate_id():
    return f"replicated_{str(uuid.uuid4())}"

def replicated_log(str):
    print(f"[REPLICATED] {str}")

def replicated_compact(caller, func, args, kwargs):
    return {
        "id" : caller.id,
        "func_name" : func.__name__,
        "args" : args,
        "kwargs" : kwargs
    }

def replicated_destroy(object):
    object_id = object.id
    object.need_destroy = True
    replicated_log(f"Requested a destroy for object with the id of : {object_id}")


class Replicator():
    def __init__(self) -> None:
        self.id = replicated_generate_id()
        self.class_name = None
        self.handler = None
        self.network_handler = None
        self.need_destroy = False
        self.replicates = []
        self.auto_datas = {}

        replicated_log(f"Object successfully replicated ! (id : {self.id})")

    def set_handler(self, new_handler):
        self.handler = new_handler

        if type(self.handler) == UrsinaNetworkingServer:
            if hasattr(self, "init_server"): self.init_server()
            
        if type(self.handler) == UrsinaNetworkingClient:
            if hasattr(self, "init_client"): self.init_client()

    def replicate(self, arg):
        self.replicates.append(arg)

    def rpc_server(self, func, *args, **kwargs):
        self.handler.send_message("replicated_sv_rpc", replicated_compact(self, func, args, kwargs))

    def rpc_multicast(self, func, *args, **kwargs):
        self.handler.broadcast("replicated_cl_rpc", replicated_compact(self, func, args, kwargs))

class ReplicatedSvEventsHandler:
    def __init__(self, server) -> None:
        self.server = server
        self.replicated_objects = {}
        self.delay = 0.01

        replicated_log(f"Handler initiated ! (delay : {self.delay}s)")

        @self.server.event
        def onClientConnected(client):
            client.send_message("replicated_cl_welcome", self.replicated_objects)

        @self.server.event
        def replicated_sv_rpc(client, datas):
            func = getattr(self.replicated_objects[datas["id"]]["server_instance"], datas["func_name"])
            func(*datas["args"], **datas["kwargs"])

    def create_replicated_object(self, class_name, *args, **kwargs):
        object_instance = class_name(*args, **kwargs)
        object_instance.class_name = class_name
        object_instance.replicated_handler = self
        object_instance.set_handler(self.server)
        added = self.replicated_objects[object_instance.id] = {
            "server_instance" : object_instance,
            "id" : object_instance.id,
            "class_name" : object_instance.class_name,
            "args" : args,
            "kwargs" : kwargs
        }
        self.server.broadcast("replicated_cl_new_object", added)
        return object_instance

    def replicated_update(self):
        r_replicates = []
        r_to_destroy = []

        for i in self.replicated_objects:
            e = self.replicated_objects[i]["server_instance"]

            if hasattr(e, "tick_server"): e.tick_server()

            if e.need_destroy == True:
                self.server.broadcast("replicated_cl_destroy", e.id)
                r_to_destroy.append(e.id)
            else:
                for replicate in e.replicates:
                    r_replicates.append({
                        "id" : e.id,
                        "name" : replicate,
                        "value" : getattr(e, replicate)
                    })
                
                #print(r_replicates)

        self.server.broadcast("replicated_cl_update", r_replicates)

        for i in r_to_destroy:
            destroy(self.replicated_objects[i]["server_instance"])
            del self.replicated_objects[i]
            replicated_log(f"Removed : {i}")

        time.sleep(self.delay)


class ReplicatedClEventsHandler:
    def __init__(self, replicated_context) -> None:
        self.client = replicated_context
        self.replicated_objects = {}
        self.auto_datas = {}
        
        @self.client.event
        def replicated_cl_welcome(datas):
            for e in datas:
                self.add_object_by_network_datas(datas[e])

        @self.client.event
        def replicated_cl_new_object(object_data):
            self.add_object_by_network_datas(object_data)

        @self.client.event
        def replicated_cl_update(datas):
            for e in datas:
                setattr(self.replicated_objects[e["id"]], e["name"], e["value"])

        @self.client.event
        def replicated_cl_rpc(datas):
            func = getattr(self.replicated_objects[datas["id"]], datas["func_name"])
            func(*datas["args"], **datas["kwargs"])

        @self.client.event
        def replicated_cl_destroy(object_id):
            object_to_delete = self.replicated_objects[object_id]
            destroy(object_to_delete)
            del self.replicated_objects[object_id]
            replicated_log(f"Removed : {object_id}")

    def add_object_by_network_datas(self, object_data):
        id = object_data["id"]
        object_instance = object_data["class_name"](*object_data["args"], **object_data["kwargs"])
        object_instance.set_handler(self.client)
        object_instance.replicated_handler = self
        object_instance.id = id
        object_instance.auto_datas = self.auto_datas
        self.replicated_objects[id] = object_instance

    def replicated_update(self):
        for e in self.replicated_objects:
            if hasattr(self.replicated_objects[e], "tick_client"):
                self.replicated_objects[e].tick_client()


from ursinanetworking import *
from easyursinanetworking import *

Client = UrsinaNetworkingClient("localhost", 25565)
Easy = EasyUrsinaNetworkingClient(Client)

@Easy.event
def onReplicatedVariableCreated(variable):
    print("Variable Created : ")
    print(variable)

@Easy.event
def onReplicatedVariableUpdated(variable):
    print("Variable Updated : ")
    print(variable)

@Easy.event
def onReplicatedVariableRemoved(variable):
    print("Variable Removed : ")
    print(variable)

while True:
    Easy.process_net_events()


"""
  _    _          _             _   _      _                      _    _
 | |  | |        (_)           | \ | |    | |                    | |  (_)
 | |  | |_ __ ___ _ _ __   __ _|  \| | ___| |___      _____  _ __| | ___ _ __   __ _
 | |  | | '__/ __| | '_ \ / _` | . ` |/ _ \ __\ \ /\ / / _ \| '__| |/ / | '_ \ / _` |
 | |__| | |  \__ \ | | | | (_| | |\  |  __/ |_ \ V  V / (_) | |  |   <| | | | | (_| |
  \____/|_|  |___/_|_| |_|\__,_|_| \_|\___|\__| \_/\_/ \___/|_|  |_|\_\_|_| |_|\__, |
                                                                                __/ |
                                                                               |___/
By K3#4869 and Squiggle#1385
"""

import socket
import threading
import pickle
import zlib

HEADERSIZE = 10
MESSAGE_LENGTH = 8
BUFFERSIZE = 4096

BUILTIN_EVENT_CONNECTION_ESTABLISHED = "onConnectionEstablished"
BUILTIN_EVENT_CONNECTION_ERROR      = "onConnectionError"

BUILTIN_EVENT_CLIENT_CONNECTED      = "onClientConnected"
BUILTIN_EVENT_CLIENT_DISCONNECTED   = "onClientDisconnected"

STATE_HEADER    = "STATE_HEADER"
STATE_PAYLOAD   = "STATE_PAYLOAD"

def ursina_networking_log(Class_, Context_, Message_):

    print(f"[{Class_} / {Context_}] {Message_}")

def ursina_networking_decompress_file(Datas_):

    return zlib.decompress(Datas_)

def ursina_networking_encode_file(Path_):

    file = open(Path_, "rb")
    datas = file.read()
    file.close()
    return zlib.compress(datas)

def ursina_networking_encode_message(Message_, Content_):

    try:
        Message = {
            "Message"   :   Message_,
            "Content"   :   Content_
        }
        EncodedMessage = pickle.dumps(Message)
        MessageLength = len(EncodedMessage)
        LengthToBytes = MessageLength.to_bytes(MESSAGE_LENGTH, byteorder = "big")
        FinalMessage = LengthToBytes + EncodedMessage
        return FinalMessage
    except Exception as e:
        ursina_networking_log("ursina_networking_encode_message", "func", e)
    return b""

class UrsinaNetworkingEvents():

    def __init__(self, lock):
        self.events = []
        self.event_table = {}
        self.lock = lock

    def push_event(self, name, *args):
        self.lock.acquire()
        self.events.append((name, args))
        self.lock.release()

    def process_net_events(self):
        self.lock.acquire()
        for event in self.events:
            Func = event[0]
            Args = event[1]
            try:
                for events_ in self.event_table:
                    for event_ in self.event_table[ events_ ]:
                        if Func in event_.__name__:
                            event_(*Args)
            except Exception as e:
                ursina_networking_log("UrsinaNetworkingEvents", "process_net_events", f"Unable to correctly call '{Func}' : '{e}'")
        self.events.clear()
        self.lock.release()

    def event(self, func):
        if func.__name__ in self.event_table:
            self.event_table[func.__name__].append(func)
        else:
            self.event_table[func.__name__]= [func]

class UrsinaNetworinkDatagramsBuffer():

    def __init__(self):
        self.header = bytes()
        self.payload = bytes()
        self.buf = bytearray()
        self.pickled_datas = None
        self.payload_length = 0
        self.receive_all = False
        self.datagrams = []
        self.state = STATE_HEADER

    def receive_datagrams(self, client_):

        self.buf += client_.recv(BUFFERSIZE)

        while True:

            self.state_changed = False

            if self.state == STATE_HEADER:

                if len(self.buf) >= MESSAGE_LENGTH:

                    self.header = self.buf[:MESSAGE_LENGTH]

                    del self.buf[:MESSAGE_LENGTH]

                    self.payload_length = int.from_bytes(self.header, byteorder = "big", signed = False)

                    self.state = STATE_PAYLOAD
                    self.state_changed = True

            elif self.state == STATE_PAYLOAD:

                if len(self.buf) >= self.payload_length:

                    self.payload = self.buf[:self.payload_length]

                    del self.buf[:self.payload_length]

                    self.state = STATE_HEADER
                    self.state_changed = True
                    self.receive_all = True
                    self.pickled_datas = pickle.loads(self.payload)
                    self.datagrams.append(self.pickled_datas)

            if not self.state_changed:
                break

    def receive(self):
        if self.receive_all:
            self.receive_all = False
            return True
        else:
            return False

class UrsinaNetworkingConnectedClient():

    def __init__(self, socket, address, id):
        self.socket = socket
        self.address = address
        self.id = id
        self.name = f"Client {id}"
        self.datas = {}

    def __repr__(self):
        return self.name

    def send_message(self, Message_, Content_):
        try:
            if (Encoded:=ursina_networking_encode_message(Message_, Content_))==b"": return False
            self.socket.sendall(Encoded)
            return True
        except Exception as e:
            ursina_networking_log("UrsinaNetworkingConnectedClient", "send_message", e)
            return False

class UrsinaNetworkingServer():

    def __init__(self, Ip_, Port_):

        self.lock = threading.Lock()
        self.events_manager = UrsinaNetworkingEvents(self.lock)
        self.network_buffer = UrsinaNetworinkDatagramsBuffer()
        self.event = self.events_manager.event
        self.clients = []
        self.lock = threading.Lock()

        try:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.bind((Ip_, Port_))
            self.server.listen()
            self.receiveThread = threading.Thread(target = self.receive)
            self.receiveThread.start()

            ursina_networking_log("UrsinaNetworkingServer", "__init__", "Server started !")
            ursina_networking_log("UrsinaNetworkingServer", "__init__", f"Ip   :   {Ip_}")
            ursina_networking_log("UrsinaNetworkingServer", "__init__", f"Port :   {Port_}")

        except Exception as e:
            ursina_networking_log("UrsinaNetworkingServer", "__init__", f"Cannot create the server : {e}")

    def process_net_events(self):
        self.events_manager.process_net_events()

    def get_client_id(self, Client_):
        for Client in self.clients:
            if Client.socket == Client_:
                return Client.id
        return None

    def get_clients_ids(self):
        Ret = []
        for Client in self.clients:
            Ret.append(Client.id)
        return Ret

    def get_client(self, Client_):
        for Client in self.clients:
            if Client.socket == Client_:
                return Client
        return None

    def get_clients(self):
        return self.clients

    def broadcast(self, Message_, Content_, IgnoreList = []):
        for Client in self.clients:
            if not Client in IgnoreList:
                Client.send_message(Message_, Content_)

    def handle(self, Client_):
        while True:
            try:
                self.network_buffer.receive_datagrams(Client_)

                for datagram in self.network_buffer.datagrams:

                    self.events_manager.push_event(datagram["Message"], self.get_client(Client_), datagram["Content"])

                self.network_buffer.datagrams = []

            except ConnectionError as e:
                ClientCopy = self.get_client(Client_)
                for Client in self.clients:
                    if Client.socket == Client_:
                        self.clients.remove(Client)
                        break

                self.events_manager.push_event(BUILTIN_EVENT_CLIENT_DISCONNECTED, ClientCopy)
                Client_.close()
                break

            except Exception as e:
                ursina_networking_log("UrsinaNetworkingServer", "handle", f"unknown error : {e}")
                break

    def receive(self):

        while True:

            client, address = self.server.accept()

            self.clients.append(UrsinaNetworkingConnectedClient(client, address, len(self.clients)))

            self.events_manager.push_event(BUILTIN_EVENT_CLIENT_CONNECTED, self.get_client(client))

            self.handle_thread = threading.Thread(target = self.handle, args = (client,))
            self.handle_thread.start()

class UrsinaNetworkingClient():

    def __init__(self, Ip_, Port_):

            try:
                self.connected = False
                self.lock = threading.Lock()
                self.events_manager = UrsinaNetworkingEvents(self.lock)
                self.network_buffer = UrsinaNetworinkDatagramsBuffer()
                self.event = self.events_manager.event
                self.connected = False
                self.handle_thread = threading.Thread(target = self.handle, args = (Ip_, Port_,))
                self.handle_thread.start()
                self.lock = threading.Lock()
            except Exception as e:
                ursina_networking_log("UrsinaNetworkingClient", "__init__", f"Cannot connect to the server : {e}")

    def process_net_events(self):
        self.events_manager.process_net_events()

    def handle(self, Ip_, Port_):
            try:
                self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.connection_response = self.client.connect_ex((Ip_, Port_))

                if self.connection_response == 0:
                    self.connected = True
                    self.events_manager.push_event(BUILTIN_EVENT_CONNECTION_ESTABLISHED)

                    ursina_networking_log("UrsinaNetworkingClient", "handle", "Client connected successfully !")

                    while True:

                        try:

                            self.network_buffer.receive_datagrams(self.client)

                            for datagram in self.network_buffer.datagrams:
                                self.events_manager.push_event(datagram["Message"], datagram["Content"])

                            self.network_buffer.datagrams = []

                        except ConnectionError as e:
                            self.events_manager.push_event(BUILTIN_EVENT_CONNECTION_ERROR, e)
                            ursina_networking_log("UrsinaNetworkingClient", "handle", f"connectionError : {e}")
                            break
                        except Exception as e:
                            ursina_networking_log("UrsinaNetworkingClient", "handle", f"unknown error : {e}")
                            break
                else:
                    self.events_manager.push_event(BUILTIN_EVENT_CONNECTION_ERROR, self.connection_response)

            except Exception as e:
                self.events_manager.push_event("connectionError", e)
                ursina_networking_log("UrsinaNetworkingClient", "handle", f"Connection Error : {e}")

    def send_message(self, Message_, Content_):
        try:
            if self.connected:
                if (Encoded:=ursina_networking_encode_message(Message_, Content_))==b"": return False
                self.client.sendall(Encoded)
                return True
            else:
                ursina_networking_log("UrsinaNetworkingClient", "send_message", f"WARNING : You are trying to send a message but the socket is not connected !")
        except Exception as e:
            ursina_networking_log("UrsinaNetworkingClient", "send_message", e)
            return False